@@ -1,18 +1,18 @@
-
-import logging
 import struct
 import re
 
+from .exif_log import get_logger
 from .utils import s2n_motorola, s2n_intel, Ratio
 from .tags import *
 
-logger = logging.getLogger('exifread')
+logger = get_logger()
 
 
 class IfdTag:
     """
     Eases dealing with tags.
     """
+
     def __init__(self, printable, tag, field_type, values, field_offset,
                  field_length):
         # printable version of data
@@ -34,14 +34,14 @@ class IfdTag:
     def __repr__(self):
         try:
             s = '(0x%04X) %s=%s @ %d' % (self.tag,
-                                        FIELD_TYPES[self.field_type][2],
-                                        self.printable,
-                                        self.field_offset)
+                                         FIELD_TYPES[self.field_type][2],
+                                         self.printable,
+                                         self.field_offset)
         except:
             s = '(%s) %s=%s @ %s' % (str(self.tag),
-                                        FIELD_TYPES[self.field_type][2],
-                                        self.printable,
-                                        str(self.field_offset))
+                                     FIELD_TYPES[self.field_type][2],
+                                     self.printable,
+                                     str(self.field_offset))
         return s
 
 
@@ -49,7 +49,9 @@ class ExifHeader:
     """
     Handle an EXIF header.
     """
-    def __init__(self, file, endian, offset, fake_exif, strict, debug=0, detailed=True):
+
+    def __init__(self, file, endian, offset, fake_exif, strict,
+                 debug=False, detailed=True):
         self.file = file
         self.endian = endian
         self.offset = offset
@@ -59,70 +61,72 @@ class ExifHeader:
         self.detailed = detailed
         self.tags = {}
 
-
     def s2n(self, offset, length, signed=0):
         """
         Convert slice to integer, based on sign and endian flags.
 
-        Usually this offset is assumed to be relative to the begining of the
+        Usually this offset is assumed to be relative to the beginning of the
         start of the EXIF information.
         For some cameras that use relative tags, this offset may be relative
         to some other starting point.
         """
         self.file.seek(self.offset + offset)
-        slice = self.file.read(length)
+        sliced = self.file.read(length)
         if self.endian == 'I':
-            val = s2n_intel(slice)
+            val = s2n_intel(sliced)
         else:
-            val = s2n_motorola(slice)
-        # Sign extension ?
+            val = s2n_motorola(sliced)
+            # Sign extension?
         if signed:
-            msb = 1 << (8*length-1)
+            msb = 1 << (8 * length - 1)
             if val & msb:
-                val = val-(msb << 1)
+                val -= (msb << 1)
         return val
 
-
     def n2s(self, offset, length):
         """Convert offset to string."""
         s = ''
         for dummy in range(length):
             if self.endian == 'I':
-                s = s + chr(offset & 0xFF)
+                s += chr(offset & 0xFF)
             else:
                 s = chr(offset & 0xFF) + s
             offset = offset >> 8
         return s
 
-
-    def first_IFD(self):
+    def _first_ifd(self):
         """Return first IFD."""
         return self.s2n(4, 4)
 
-
-    def next_IFD(self, ifd):
+    def _next_ifd(self, ifd):
         """Return the pointer to next IFD."""
         entries = self.s2n(ifd, 2)
-        next_ifd = self.s2n(ifd+2+12*entries, 4)
+        next_ifd = self.s2n(ifd + 2 + 12 * entries, 4)
         if next_ifd == ifd:
             return 0
         else:
             return next_ifd
 
-
-    def list_IFDs(self):
+    def list_ifd(self):
         """Return the list of IFDs in the header."""
-        i = self.first_IFD()
+        i = self._first_ifd()
         ifds = []
         while i:
             ifds.append(i)
-            i = self.next_IFD(i)
+            i = self._next_ifd(i)
         return ifds
 
+    def dump_ifd(self, ifd, ifd_name, tag_dict=EXIF_TAGS, relative=0, stop_tag=DEFAULT_STOP_TAG):
+        """
+        Return a list of entries in the given IFD.
+        """
+        # make sure we can process the entries
+        try:
+            entries = self.s2n(ifd, 2)
+        except TypeError:
+            logger.warning("Possibly corrupted IFD: %s" % ifd)
+            return
 
-    def dump_IFD(self, ifd, ifd_name, tag_dict=EXIF_TAGS, relative=0, stop_tag=DEFAULT_STOP_TAG):
-        """Return a list of entries in the given IFD."""
-        entries = self.s2n(ifd, 2)
         for i in range(entries):
             # entry is index of start of this IFD in the file
             entry = ifd + 2 + 12 * i
@@ -144,9 +148,9 @@ class ExifHeader:
                     if not self.strict:
                         continue
                     else:
-                        raise ValueError('unknown type %d in tag 0x%04X' % (field_type, tag))
+                        raise ValueError('Unknown type %d in tag 0x%04X' % (field_type, tag))
 
-                typelen = FIELD_TYPES[field_type][0]
+                type_length = FIELD_TYPES[field_type][0]
                 count = self.s2n(entry + 4, 4)
                 # Adjust for tag id/type/count (2+2+4 bytes)
                 # Now we point at either the data or the 2nd level offset
@@ -154,7 +158,7 @@ class ExifHeader:
 
                 # If the value fits in 4 bytes, it is inlined, else we
                 # need to jump ahead again.
-                if count * typelen > 4:
+                if count * type_length > 4:
                     # offset is not the value; it's a pointer to the value
                     # if relative we set things up so s2n will seek to the right
                     # place when it adds self.offset.  Note that this 'relative'
@@ -165,7 +169,7 @@ class ExifHeader:
                         tmp_offset = self.s2n(offset, 4)
                         offset = tmp_offset + ifd - 8
                         if self.fake_exif:
-                            offset = offset + 18
+                            offset += 18
                     else:
                         offset = self.s2n(offset, 4)
 
@@ -175,14 +179,24 @@ class ExifHeader:
                     # special case: null-terminated ASCII string
                     # XXX investigate
                     # sometimes gets too big to fit in int value
-                    if count != 0: # and count < (2**31):  # 2E31 is hardware dependant. --gd
+                    if count != 0:  # and count < (2**31):  # 2E31 is hardware dependant. --gd
+                        file_position = self.offset + offset
                         try:
-                            self.file.seek(self.offset + offset)
+                            self.file.seek(file_position)
                             values = self.file.read(count)
-                            #print values
+                            #print(values)
                             # Drop any garbage after a null.
-                            values = values.split('\x00', 1)[0]
+                            values = values.split(b'\x00', 1)[0]
+                            if isinstance(values, bytes):
+                                try:
+                                    values = values.decode("utf-8")
+                                except UnicodeDecodeError:
+                                    logger.warning("Possibly corrupted field %s in %s IFD", tag_name, ifd_name)
                         except OverflowError:
+                            logger.warn('OverflowError at position: %s, length: %s', file_position, count)
+                            values = ''
+                        except MemoryError:
+                            logger.warn('MemoryError at position: %s, length: %s', file_position, count)
                             values = ''
                 else:
                     values = []
@@ -196,37 +210,45 @@ class ExifHeader:
                             if field_type in (5, 10):
                                 # a ratio
                                 value = Ratio(self.s2n(offset, 4, signed),
-                                                self.s2n(offset + 4, 4, signed))
+                                              self.s2n(offset + 4, 4, signed))
                             else:
-                                value = self.s2n(offset, typelen, signed)
+                                value = self.s2n(offset, type_length, signed)
                             values.append(value)
-                            offset = offset + typelen
+                            offset = offset + type_length
                     # The test above causes problems with tags that are
-                    # supposed to have long values!  Fix up one important case.
-                    elif tag_name in ('MakerNote',
-                                      makernote.canon.CAMERA_INFO_TAG_NAME):
+                    # supposed to have long values! Fix up one important case.
+                    elif tag_name in ('MakerNote', makernote.canon.CAMERA_INFO_TAG_NAME):
                         for dummy in range(count):
-                            value = self.s2n(offset, typelen, signed)
+                            value = self.s2n(offset, type_length, signed)
                             values.append(value)
-                            offset = offset + typelen
-                    #else :
-                    #    print "Warning: dropping large tag:", tag, tag_name
+                            offset = offset + type_length
 
                 # now 'values' is either a string or an array
                 if count == 1 and field_type != 2:
                     printable = str(values[0])
-                elif count > 50 and len(values) > 20 :
-                    printable = str( values[0:20] )[0:-1] + ", ... ]"
+                elif count > 50 and len(values) > 20:
+                    printable = str(values[0:20])[0:-1] + ", ... ]"
                 else:
-                    printable = str(values)
+                    try:
+                        printable = str(values)
+                    # fix for python2's handling of unicode values
+                    except UnicodeEncodeError:
+                        printable = unicode(values)
 
                 # compute printable version of values
                 if tag_entry:
+                    # optional 2nd tag element is present
                     if len(tag_entry) != 1:
-                        # optional 2nd tag element is present
                         if callable(tag_entry[1]):
                             # call mapping function
                             printable = tag_entry[1](values)
+                        elif type(tag_entry[1]) is tuple:
+                            ifd_info = tag_entry[1]
+                            try:
+                                logger.debug('%s SubIFD at offset %d:', ifd_info[0], values[0])
+                                self.dump_ifd(values[0], ifd_info[0], tag_dict=ifd_info[1], stop_tag=stop_tag)
+                            except IndexError:
+                                logger.warn('No values found for %s SubIFD', ifd_info[0])
                         else:
                             printable = ''
                             for i in values:
@@ -234,15 +256,19 @@ class ExifHeader:
                                 printable += tag_entry[1].get(i, repr(i))
 
                 self.tags[ifd_name + ' ' + tag_name] = IfdTag(printable, tag,
-                                                            field_type,
-                                                            values, field_offset,
-                                                            count * typelen)
-                logger.debug(" %s: %s", tag_name, repr(self.tags[ifd_name + ' ' + tag_name]))
+                                                              field_type,
+                                                              values, field_offset,
+                                                              count * type_length)
+                try:
+                    tag_value = repr(self.tags[ifd_name + ' ' + tag_name])
+                # fix for python2's handling of unicode values
+                except UnicodeEncodeError:
+                    tag_value = unicode(self.tags[ifd_name + ' ' + tag_name])
+                logger.debug(' %s: %s', tag_name, tag_value)
 
             if tag_name == stop_tag:
                 break
 
-
     def extract_tiff_thumbnail(self, thumb_ifd):
         """
         Extract uncompressed TIFF thumbnail.
@@ -260,37 +286,37 @@ class ExifHeader:
             tiff = 'MM\x00*\x00\x00\x00\x08'
         else:
             tiff = 'II*\x00\x08\x00\x00\x00'
-        # ... plus thumbnail IFD data plus a null "next IFD" pointer
+            # ... plus thumbnail IFD data plus a null "next IFD" pointer
         self.file.seek(self.offset + thumb_ifd)
-        tiff += self.file.read(entries*12+2) + '\x00\x00\x00\x00'
+        tiff += self.file.read(entries * 12 + 2) + '\x00\x00\x00\x00'
 
         # fix up large value offset pointers into data area
         for i in range(entries):
             entry = thumb_ifd + 2 + 12 * i
             tag = self.s2n(entry, 2)
-            field_type = self.s2n(entry+2, 2)
-            typelen = FIELD_TYPES[field_type][0]
-            count = self.s2n(entry+4, 4)
-            oldoff = self.s2n(entry+8, 4)
+            field_type = self.s2n(entry + 2, 2)
+            type_length = FIELD_TYPES[field_type][0]
+            count = self.s2n(entry + 4, 4)
+            old_offset = self.s2n(entry + 8, 4)
             # start of the 4-byte pointer area in entry
             ptr = i * 12 + 18
             # remember strip offsets location
             if tag == 0x0111:
                 strip_off = ptr
-                strip_len = count * typelen
-            # is it in the data area?
-            if count * typelen > 4:
+                strip_len = count * type_length
+                # is it in the data area?
+            if count * type_length > 4:
                 # update offset pointer (nasty "strings are immutable" crap)
                 # should be able to say "tiff[ptr:ptr+4]=newoff"
                 newoff = len(tiff)
-                tiff = tiff[:ptr] + self.n2s(newoff, 4) + tiff[ptr+4:]
+                tiff = tiff[:ptr] + self.n2s(newoff, 4) + tiff[ptr + 4:]
                 # remember strip offsets location
                 if tag == 0x0111:
                     strip_off = newoff
                     strip_len = 4
                 # get original data and store it
-                self.file.seek(self.offset + oldoff)
-                tiff += self.file.read(count * typelen)
+                self.file.seek(self.offset + old_offset)
+                tiff += self.file.read(count * type_length)
 
         # add pixel strips and update strip offset info
         old_offsets = self.tags['Thumbnail StripOffsets'].values
@@ -306,7 +332,6 @@ class ExifHeader:
 
         self.tags['TIFFThumbnail'] = tiff
 
-
     def extract_jpeg_thumbnail(self):
         """
         Extract JPEG thumbnail.
@@ -327,7 +352,6 @@ class ExifHeader:
                 self.file.seek(self.offset + thumb_offset.values[0])
                 self.tags['JPEGThumbnail'] = self.file.read(thumb_offset.field_length)
 
-
     def decode_maker_note(self):
         """
         Decode all the camera-specific MakerNote formats
@@ -354,12 +378,7 @@ class ExifHeader:
 
         # Some apps use MakerNote tags but do not use a format for which we
         # have a description, so just do a raw dump for these.
-        #if self.tags.has_key('Image Make'):
         make = self.tags['Image Make'].printable
-        #else:
-        #    make = ''
-
-        # model = self.tags['Image Model'].printable # unused
 
         # Nikon
         # The maker note usually starts with the word Nikon, followed by the
@@ -369,27 +388,26 @@ class ExifHeader:
         if 'NIKON' in make:
             if note.values[0:7] == [78, 105, 107, 111, 110, 0, 1]:
                 logger.debug("Looks like a type 1 Nikon MakerNote.")
-                self.dump_IFD(note.field_offset+8, 'MakerNote',
-                              tag_dict=makernote.NIKON_OLD)
+                self.dump_ifd(note.field_offset + 8, 'MakerNote',
+                              tag_dict=makernote.nikon.TAGS_OLD)
             elif note.values[0:7] == [78, 105, 107, 111, 110, 0, 2]:
-                if self.debug:
-                    logger.debug("Looks like a labeled type 2 Nikon MakerNote")
+                logger.debug("Looks like a labeled type 2 Nikon MakerNote")
                 if note.values[12:14] != [0, 42] and note.values[12:14] != [42, 0]:
                     raise ValueError("Missing marker tag '42' in MakerNote.")
-                # skip the Makernote label and the TIFF header
-                self.dump_IFD(note.field_offset+10+8, 'MakerNote',
-                              tag_dict=makernote.NIKON_NEW, relative=1)
+                    # skip the Makernote label and the TIFF header
+                self.dump_ifd(note.field_offset + 10 + 8, 'MakerNote',
+                              tag_dict=makernote.nikon.TAGS_NEW, relative=1)
             else:
                 # E99x or D1
                 logger.debug("Looks like an unlabeled type 2 Nikon MakerNote")
-                self.dump_IFD(note.field_offset, 'MakerNote',
-                              tag_dict=makernote.NIKON_NEW)
+                self.dump_ifd(note.field_offset, 'MakerNote',
+                              tag_dict=makernote.nikon.TAGS_NEW)
             return
 
         # Olympus
         if make.startswith('OLYMPUS'):
-            self.dump_IFD(note.field_offset+8, 'MakerNote',
-                          tag_dict=makernote.OLYMPUS)
+            self.dump_ifd(note.field_offset + 8, 'MakerNote',
+                          tag_dict=makernote.olympus.TAGS)
             # TODO
             #for i in (('MakerNote Tag 0x2020', makernote.OLYMPUS_TAG_0x2020),):
             #    self.decode_olympus_tag(self.tags[i[0]].values, i[1])
@@ -397,8 +415,8 @@ class ExifHeader:
 
         # Casio
         if 'CASIO' in make or 'Casio' in make:
-            self.dump_IFD(note.field_offset, 'MakerNote',
-                          tag_dict=makernote.CASIO)
+            self.dump_ifd(note.field_offset, 'MakerNote',
+                          tag_dict=makernote.casio.TAGS)
             return
 
         # Fujifilm
@@ -412,16 +430,27 @@ class ExifHeader:
             offset = self.offset
             self.offset += note.field_offset
             # process note with bogus values (note is actually at offset 12)
-            self.dump_IFD(12, 'MakerNote', tag_dict=makernote.FUJIFILM)
+            self.dump_ifd(12, 'MakerNote', tag_dict=makernote.fujifilm.TAGS)
             # reset to correct values
             self.endian = endian
             self.offset = offset
             return
 
+        # Apple
+        if make == 'Apple' and \
+                note.values[0:10] == [65, 112, 112, 108, 101, 32, 105, 79, 83, 0]:
+            t = self.offset
+            self.offset += note.field_offset+14
+            self.dump_ifd(0, 'MakerNote',
+                          tag_dict=makernote.apple.TAGS)
+            self.offset = t
+            return
+
         # Canon
         if make == 'Canon':
-            self.dump_IFD(note.field_offset, 'MakerNote',
+            self.dump_ifd(note.field_offset, 'MakerNote',
                           tag_dict=makernote.canon.TAGS)
+
             for i in (('MakerNote Tag 0x0001', makernote.canon.CAMERA_SETTINGS),
                       ('MakerNote Tag 0x0002', makernote.canon.FOCAL_LENGTH),
                       ('MakerNote Tag 0x0004', makernote.canon.SHOT_INFO),
@@ -429,22 +458,20 @@ class ExifHeader:
                       ('MakerNote Tag 0x0093', makernote.canon.FILE_INFO)):
                 if i[0] in self.tags:
                     logger.debug('Canon ' + i[0])
-                    self.canon_decode_tag(self.tags[i[0]].values, i[1])
+                    self._canon_decode_tag(self.tags[i[0]].values, i[1])
                     del self.tags[i[0]]
             if makernote.canon.CAMERA_INFO_TAG_NAME in self.tags:
                 tag = self.tags[makernote.canon.CAMERA_INFO_TAG_NAME]
                 logger.debug('Canon CameraInfo')
-                self.canon_decode_camera_info(tag)
+                self._canon_decode_camera_info(tag)
                 del self.tags[makernote.canon.CAMERA_INFO_TAG_NAME]
             return
 
-
-    def olympus_decode_tag(self, value, mn_tags):
+    def _olympus_decode_tag(self, value, mn_tags):
         """ TODO Decode Olympus MakerNote tag based on offset within tag."""
         pass
 
-
-    def canon_decode_tag(self, value, mn_tags):
+    def _canon_decode_tag(self, value, mn_tags):
         """
         Decode Canon MakerNote tag based on offset within tag.
 
@@ -461,22 +488,23 @@ class ExifHeader:
                 logger.debug(" %s %s %s", i, name, hex(value[i]))
             except TypeError:
                 logger.debug(" %s %s %s", i, name, value[i])
-            
+
             # it's not a real IFD Tag but we fake one to make everybody
             # happy. this will have a "proprietary" type
             self.tags['MakerNote ' + name] = IfdTag(str(val), None, 0, None,
                                                     None, None)
 
-    def canon_decode_camera_info(self, camera_info_tag):
-        """Decode the variable length encoded camera info section."""
+    def _canon_decode_camera_info(self, camera_info_tag):
+        """
+        Decode the variable length encoded camera info section.
+        """
         model = self.tags.get('Image Model', None)
         if not model:
             return
         model = str(model.values)
 
         camera_info_tags = None
-        for (model_name_re, tag_desc) in \
-              makernote.canon.CAMERA_INFO_MODEL_MAP.items():
+        for (model_name_re, tag_desc) in makernote.canon.CAMERA_INFO_MODEL_MAP.items():
             if re.search(model_name_re, model):
                 camera_info_tags = tag_desc
                 break
@@ -496,7 +524,7 @@ class ExifHeader:
             tag_size = struct.calcsize(tag_format)
             if len(camera_info) < offset + tag_size:
                 continue
-            packed_tag_value = camera_info[offset:offset+tag_size]
+            packed_tag_value = camera_info[offset:offset + tag_size]
             tag_value = struct.unpack(tag_format, packed_tag_value)[0]
 
             tag_name = tag[0]
@@ -509,3 +537,17 @@ class ExifHeader:
 
             self.tags['MakerNote ' + tag_name] = IfdTag(str(tag_value), None,
                                                         0, None, None, None)
+
+    def parse_xmp(self, xmp_string):
+        import xml.dom.minidom
+
+        logger.debug('XMP cleaning data')
+
+        xml = xml.dom.minidom.parseString(xmp_string)
+        pretty = xml.toprettyxml()
+        cleaned = []
+        for line in pretty.splitlines():
+            if line.strip():
+                cleaned.append(line)
+        self.tags['Image ApplicationNotes'] = IfdTag('\n'.join(cleaned), None,
+                                                     1, None, None, None)
